﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Data;
using System.Windows.Documents;
using System.Windows.Input;
using System.Windows.Media;
using System.Windows.Media.Imaging;
using System.Windows.Navigation;
using System.IO;
using WMPLib;
using System.Timers;
using Musiclibrary.Properties;
using System.ComponentModel;
using Microsoft.WindowsAPICodePack.Taskbar;
using System.Reflection;
using System.Threading;
using System.Collections.ObjectModel;
using Microsoft.WindowsAPICodePack.Dialogs;

namespace Musiclibrary {
    /// <summary>
    /// Interaction logic for Window1.xaml
    /// </summary>
    public partial class MainWindow : Window {
   
        protected System.Timers.Timer PositionUpdateTimer;
        protected Win7Tools Win7Tools;
        protected DialogAdorner Adorner;

        public ObservableCollection<IPlaylist> Playlists { get; set; }

        #region Initialization

        public MainWindow() {
            InitializeComponent();
            Instance = this;
            this.Playlists = new ObservableCollection<IPlaylist>();
        }

        public static MainWindow Instance { get; protected set; }

        private void Window_Loaded(object sender, RoutedEventArgs e) {
            try {
                App.This.Load();
            } catch {
                Extensions.ShowErrorBox("Sh*t! The library could not be loaded. You'll have to create a new one or open another library.");
            }

            //Load playlist source, which allows grouping
            (this.FindResource("SourcesCVS") as CollectionViewSource).Source = this.Playlists;

            //Set player event handlers
            App.This.Player.StateChanged += new Action<IPlayer, EventArgs>(Player_StateChanged);
            App.This.Player.SongChanged += new Action<IPlayer, EventArgs>(Player_SongChanged);
            App.This.Player.SongIndexChanged += new Action<IPlayer, EventArgs>(Player_SongIndexChanged);

            //Initialize timer to update the slider for the current position in the song
            this.PositionUpdateTimer = new System.Timers.Timer(40);
            this.PositionUpdateTimer.Elapsed += new ElapsedEventHandler(Position_Update);
            this.PositionUpdateTimer.Start();

            //Make player controllable from taskbar in Win7
            this.Win7Tools = new Win7Tools(this);

            //Show welcome screen if no library is loaded
            if (App.This.Library == null)
                this.ShowPopup(new Welcome(), true);
            else {
                this.LoadLibrary();

                //Automatically play playlist by command line parameter - needed for jumplist
                var args = Environment.GetCommandLineArgs();
                if (args.Length == 3 && args[1] == "-playlist") {
                    int id;
                    if (int.TryParse(args[2], out id)) {
                        var result = from i in App.This.Library.Playlists where i.Id == id select i;

                        if (result.Count() > 0) {
                            var playlist = result.Single();
                            this.PlayList(playlist);
                        }
                    }
                }
            }
        }

        #endregion

        #region Helper functions

        /// <summary>
        /// Shows the specified popup in an overlay. If a popup is already displayed, 
        /// the next popup is shown with a slide animation.
        /// </summary>
        /// <param name="popup"></param>
        /// <param name="canCancel">Can the user hide the popup by clicking in an empty area?</param>
        protected void ShowPopup(Popup popup, bool canCancel)
        {
            if (this.Adorner == null || !this.Adorner.IsShown)
            {
                this.Adorner = new DialogAdorner(this.MainGrid, popup, canCancel);
                this.Adorner.Show();
            }
            else
                this.Adorner.ShowNextDialog(popup, canCancel);
        }

        #endregion

        #region Controller functions

        /// <summary>
        /// Creates a new library at the default location and loads it to the UI.
        /// </summary>
        public void CreateLibrary() {
            App.This.CreateLibrary();
            this.LoadLibrary();
        }

        /// <summary>
        /// Loads the data from the current music library to the UI.
        /// </summary>
        public void LoadLibrary() {
            this.LoadPlaylists();
            this.LoadSongs();
            this.Win7Tools.LoadPlaylists();
        }

        /// <summary>
        /// Loads the songs from the library into the ListView.
        /// </summary>
        public void LoadSongs() {
            try {
                this.Songs.Items.Clear();
                var playlist = (this.PlaylistsList.SelectedItem as IPlaylist);

                if (playlist == null) return;

                var songs = playlist.Songs;

                var search = this.Search.Text.Trim();
                if (search.Length > 0)
                    songs = songs.Filter(search);

                foreach (var song in songs)
                    this.Songs.Items.Add(song);

                var contextMenu = this.FindResource("SongContextMenu") as ContextMenu;
                var menuItem = (from MenuItem i in contextMenu.Items where i.Name == "DeleteFromList" select i).First();
                menuItem.Visibility = playlist.ItemsEditable ? Visibility.Visible : Visibility.Collapsed;
            } catch (AccessViolationException) {
                this.LoadSongs(); //SQL Compact weirdness - just retry and it works
            } catch {
                Extensions.ShowErrorBox("The songs could not be loaded.. :-(");
            }
        }

        /// <summary>
        /// Loads the playlists from the library into the list at the left side.
        /// </summary>
        public void LoadPlaylists() {
            try {
                this.Playlists.Clear();

                this.Playlists.Add(new AllSongs());
                this.Playlists.Add(new CurrentPlaylist());

                foreach (var playlist in App.This.Library.Playlists)
                    this.Playlists.Add(playlist);

                foreach (var remoteLibrary in App.This.RemoteLibraries) {
                    this.Playlists.Add(new AllSongs() { Library = remoteLibrary.Value });
                    foreach (var playlist in remoteLibrary.Value.Playlists)
                        this.Playlists.Add(playlist);
                }

                this.PlaylistsList.SelectedIndex = 0;
            } catch (AccessViolationException) {
                this.LoadPlaylists();
            } catch {
                Extensions.ShowErrorBox("The playlists could not be loaded.. :-(");
            }
        }

        /// <summary>
        /// Plays the specified song and sets the current playlisit to the currently visible songs.
        /// </summary>
        /// <param name="song"></param>
        public void PlaySong(ISong song)
        {
            App.This.Player.Playlist.Clear();
            App.This.Player.Playlist.AddRange(this.Songs.Items.Cast<ISong>());
            App.This.Player.SongIndex = App.This.Player.Playlist.IndexOf(song);

            if (this.PlaylistsList.SelectedItem is CurrentPlaylist)
                this.LoadSongs();

            App.This.Player.Play();
        }

        /// <summary>
        /// Plays the songs from the specified source and sets this source as the current playlist.
        /// </summary>
        /// <param name="source"></param>
        public void PlayList(IPlaylist playlist)
        {
            try {
                if (!(playlist is CurrentPlaylist)) {
                    App.This.Player.Playlist.Clear();
                    App.This.Player.Playlist.AddRange(playlist.Songs);
                }
                if (App.This.Player.Playlist.Count > 0) {
                    App.This.Player.SongIndex = 0;
                    App.This.Player.Play();
                }
            } catch {
                Extensions.ShowErrorBox("Well.. actually.. I have to admit.. this didn't work.");
            }
        }

        public void PlayOrPause() {
            if (App.This.Player.State == PlayerState.Playing)
                App.This.Player.Pause();
            else
                this.Play();
        }

        public void Play() {
            if (App.This.Player.State == PlayerState.Paused)
                App.This.Player.Play();
            else {
                if (App.This.Player.Song == null && this.Songs.SelectedItem != null)
                    this.PlaySong(this.Songs.SelectedItem as ISong);
                App.This.Player.Play();
            }
        }

        #endregion

        #region Controller functions that display dialogs

        /// <summary>
        /// Imports the specified directories and files into the current library and
        /// shows a popup with the import progress.
        /// </summary>
        public void Import(IEnumerable<FileSystemInfo> filesAndDirs)
        {
            var importer = new Importer(this);
            importer.FilesAndDirs.AddRange(filesAndDirs);

            importer.Done += new EventHandler<WorkerCompletedEventArgs>((sender, e) =>
            {
                this.ShowPopup(new ImportFinished(e.Result as ImportResult), true);
                this.LoadSongs();
            });

            importer.Cancelled += new EventHandler<EventArgs>((sender, e) =>
            {
                this.Adorner.Hide();
                this.LoadSongs();
            });

            this.ShowPopup(new Working(importer, "Importing...", "Your music is being added to the library"), false);

            importer.Begin();
        }

        /// <summary>
        /// Shows a dialog to let the user choose a directory to import.
        /// </summary>
        public void ImportDir()
        {
            var dialog = new System.Windows.Forms.FolderBrowserDialog();
          
            dialog.Description = "Choose the folder you want to import:";

            if (dialog.ShowDialog() == System.Windows.Forms.DialogResult.OK)
                this.Import(new List<FileSystemInfo>() { new DirectoryInfo(dialog.SelectedPath) });
        }

        /// <summary>
        /// Shows a dialog to let the user choose a file to import.
        /// </summary>
        public void ImportSingleFile()
        {
            var dialog = new Microsoft.Win32.OpenFileDialog();
            dialog.Filter = "MP3-files (*.mp3)|*.mp3|All files (*.*)|*";

            if (dialog.ShowDialog().Value)
                this.Import(new List<FileSystemInfo>() { new FileInfo(dialog.FileName) });
        }

        /// <summary>
        /// Exports the current library to a place chosen by the user.
        /// </summary>
        public void ExportLibrary()
        {
            var dialog = new Microsoft.Win32.SaveFileDialog();
            dialog.Filter = "Music libraries (*.sdf)|*.sdf|All files (*.*)|*";

            if (dialog.ShowDialog().Value)
                try {
                    File.Copy(Properties.Settings.Default.MediaLibrary, dialog.FileName, true);
                } catch (Exception ex) {
                    Extensions.ShowErrorBox("I'm sorry, something went wrong: " + ex.Message);
                }
        }

        /// <summary>
        /// Shows a dialog to let the user choose a library to open.
        /// </summary>
        /// <returns>Returns true if the user clicked OK and the library was loaded successfully.</returns>
        public bool OpenLibrary() {
            var dialog = new Microsoft.Win32.OpenFileDialog();
            dialog.Title = "Open existing library";
            dialog.Filter = "Music libraries (*.sdf)|*.sdf|All files (*.*)|*";

            if (dialog.ShowDialog().Value) {
                try {
                    Mouse.OverrideCursor = Cursors.Wait;

                    App.This.LoadLibrary(dialog.FileName);
                    Thread.Sleep(100); //SQL Compact is weird. Without this, loading fails.
                    this.LoadLibrary();

                    Mouse.OverrideCursor = null;
                } catch (Exception ex) {
                    Extensions.ShowErrorBox("Oops! The selected library could not be loaded: " + ex.Message);

                    //Try to load previous library
                    try {
                        App.This.LoadLibrary(Properties.Settings.Default.MediaLibrary);
                        this.LoadLibrary();
                    } catch {
                        this.OpenLibrary(); //If this fails, too, show dialog again.
                    }
                }
                return true;
            } else
                return false;
        }

        #endregion

        #region Drag & drop songs to a playlist

        private List<ISong> DraggedSongs = null;

        private void Songs_PreviewMouseDown(object sender, MouseButtonEventArgs e) {
            var pointedItem = ((ListViewItem)sender).Content as ISong;
            var selectedItems = new List<ISong>(this.Songs.SelectedItems.Cast<ISong>());

            if (selectedItems.Contains(pointedItem))
                this.DraggedSongs = selectedItems;
            else
                this.DraggedSongs = new List<ISong>() { pointedItem };
        }

        private void Songs_PreviewMouseMove(object sender, MouseEventArgs e) {
            if (this.DraggedSongs != null && e.LeftButton == MouseButtonState.Pressed) {
                DataObject data = new DataObject();
                data.SetData(this.DraggedSongs);
                DragDrop.DoDragDrop(this.Songs, this.DraggedSongs, DragDropEffects.Copy);
            }
        }

        private void Playlists_Drop(object sender, DragEventArgs e) {
            var playlist = this.PlaylistsList.GetItemFromEventSender<IPlaylist>(sender);

            if (e.Data.GetDataPresent(typeof(List<ISong>))){
                var songs = e.Data.GetData(typeof(List<ISong>)) as List<ISong>;
                foreach (var song in songs) {
                    if (song is Song) //Local songs can be added to any playlist
                        playlist.AddSong(song);
                    //Remote songs cannot be added to local playlists
                    else if (song is RemoteSong && !(playlist is Playlist)) 
                        playlist.AddSong(song);
                }

                this.UpdatePrevNext();
            }
        }

        #endregion

        #region Basic event handlers 

        private void PlayPause_Click(object sender, EventArgs e)
        {
            this.PlayOrPause();
        }

        private void Previous_Click(object sender, EventArgs e)
        {
            App.This.Player.Previous();
        }

        private void Next_Click(object sender, EventArgs e)
        {
            App.This.Player.Next();
        }

        private void ImportFile_Click(object sender, RoutedEventArgs e)
        {
            this.ImportSingleFile();
        }

        private void ImportFolder_Click(object sender, RoutedEventArgs e)
        {
            this.ImportDir();
        }

        private void Share_Checked(object sender, System.Windows.RoutedEventArgs e) {
            App.This.SharingServer.Start();
        }

        private void Share_Unchecked(object sender, System.Windows.RoutedEventArgs e) {
            App.This.SharingServer.Stop();
        }

        private void FindLibraries_Click(object sender, RoutedEventArgs e)
        {
            this.ShowPopup(new FindLibraries(), true);
        }

        private void OpenLibrary_Click(object sender, RoutedEventArgs e)
        {
            this.OpenLibrary();
        }

        private void ExportLibrary_Click(object sender, RoutedEventArgs e)
        {
            this.ExportLibrary();
        }

        private void Settings_Click(object sender, System.Windows.RoutedEventArgs e)
        {
            this.ShowPopup(new Settings(), true);
        }

        private void Shuffle_Click(object sender, System.Windows.RoutedEventArgs e) {
            App.This.Player.Shuffle();
            this.LoadSongs();
        }

        private void Repeat_Click(object sender, System.Windows.RoutedEventArgs e) {
            App.This.Player.Repeat = this.Repeat.IsChecked.Value;
        }

        private void Song_DoubleClick(object sender, EventArgs e)
        {
            var song = (ISong)((ListViewItem)sender).Content;
            this.PlaySong(song);
        }
		
		private void Search_Click(object sender, System.Windows.RoutedEventArgs e)
        {
        	this.Search.Focus();
			this.Search.SelectAll();
        }

        private void Search_TextChanged(object sender, TextChangedEventArgs e)
        {
            this.LoadSongs();
        }

        private void Delete_Click(object sender, RoutedEventArgs e)
        {
            var playlist = this.PlaylistsList.GetItemFromEventSender<Playlist>(sender);

            if (playlist != null && playlist.IsEditable)
            {
                playlist.Delete();
                this.Playlists.Remove(playlist);
            }
        }

        private void Volume_ValueChanged(object sender, RoutedPropertyChangedEventArgs<double> e)
        {
            if (App.This.Player != null)
                App.This.Player.Volume = this.Volume.Value;
        }

        private void AddItemButton_ItemAdded(Musiclibrary.AddItemButton arg1, System.EventArgs arg2)
        {
            var playlist = new Playlist();
            playlist.Name = arg1.Value.Trim();

            if (playlist.Name.Length > 0)
            {
                try {
                    App.This.Library.Playlists.InsertOnSubmit(playlist);
                    App.This.Library.SubmitChanges();
                } catch {
                    Extensions.ShowErrorBox("This didn't work. Really. It just FAILED.");
                }

                this.Playlists.Add(playlist);
                this.Win7Tools.LoadPlaylists();
            }
        }

        /// <summary>
        /// Raised when double-clicking on playlist
        /// </summary>
        private void Playlist_Play(object sender, EventArgs e)
        {
            var item = sender as ListBoxItem;
            if (item != null)
            {
                var playlist = item.Content as IPlaylist;
                if (playlist != null)
                    this.PlayList(playlist);
            }
        }


        private void Window_Closed(object sender, EventArgs e) {
            App.This.Shutdown();
        }

        #endregion

        #region Position updating

        /// <summary>
        /// If this variable is set to true, the position is currently automatically updated
        /// by the position update timer.
        /// </summary>
        bool autoPositionUpdate;

        //Update the value of the slider showing the position
        private void Position_Update(object sender, ElapsedEventArgs e)
        {
            this.Dispatcher.Invoke(new Action(() =>
            {
                if (App.This.Player.State == PlayerState.Playing)
                {
                    this.Elapsed.Text = App.This.Player.Position.ToShortString();
                    this.Remaining.Text = (App.This.Player.Duration - App.This.Player.Position).ToShortString();

                    autoPositionUpdate = true;
                    this.Position.Maximum = App.This.Player.Duration.TotalSeconds;
                    autoPositionUpdate = false;

                    autoPositionUpdate = true;
                    this.Position.Value = App.This.Player.Position.TotalSeconds;
                    autoPositionUpdate = false;
                }
            }), new object[] { });
        }

        private void Position_ValueChanged(object sender, RoutedPropertyChangedEventArgs<double> e)
        {
            //If the user has changed the value, stop automatic updating until the user finishes
            if (!autoPositionUpdate)
                this.PositionUpdateTimer.Stop();
        }

        private void Position_MouseUp(object sender, MouseButtonEventArgs e)
        {
            //User has finished!
            App.This.Player.Position = new TimeSpan(0, 0, (int)this.Position.Value);
            this.PositionUpdateTimer.Start();
        }

        private void Position_MouseLeave(object sender, MouseEventArgs e)
        {
            //Also finish updating if the user takes the mouse out of the slider -
            //but only when he has started changing the value before!
            if (!this.PositionUpdateTimer.Enabled)
            {
                App.This.Player.Position = new TimeSpan(0, 0, (int)this.Position.Value);
                this.PositionUpdateTimer.Start();
            }
        }

        #endregion

        #region Song context menu event handlers

        private void DeleteCompletely_Click(object sender, RoutedEventArgs e)
        {
            var items = this.Songs.SelectedItems.Cast<ISong>().ToList();

            foreach (var item in items)
            {
                var song = item as Song;

                if (song != null) {
                    App.This.Library.Songs.DeleteOnSubmit(song);
                    this.Songs.Items.Remove(item);
                    File.Delete(song.Path);

                    App.This.Player.Playlist.Remove(song);
                }
            }

            App.This.Library.SubmitChanges();
        }

        private void DeleteFromLibrary_Click(object sender, RoutedEventArgs e)
        {
            var items = this.Songs.SelectedItems.Cast<ISong>().ToList();

            foreach (var item in items)
            {
                var song = item as Song;

                if (song != null) {
                    App.This.Library.Songs.DeleteOnSubmit(song);
                    this.Songs.Items.Remove(item);

                    App.This.Player.Playlist.Remove(song);
                }
            }

            App.This.Library.SubmitChanges();
        }

        private void DeleteFromList_Click(object sender, RoutedEventArgs e)
        {
            var playlist = this.PlaylistsList.SelectedItem as IPlaylist;

            if (playlist.ItemsEditable)
                foreach (var song in this.Songs.SelectedItems.Cast<ISong>().ToList())
                {
                    playlist.DeleteSong(song);
                    this.Songs.Items.Remove(song);
                }

            App.This.Library.SubmitChanges();
        }

        private void EditMetadata_Click(object sender, RoutedEventArgs e)
        {
            //You can only edit local songs
            var song = this.Songs.SelectedItem as Song;

            if (song != null)
                this.ShowPopup(new EditMetadata(song), true);
        }

        private void ShowInExplorer_Click(object sender, RoutedEventArgs e)
        {
            //Works only for local songs!
            var song = this.Songs.SelectedItem as Song;

            if (song != null)
                System.Diagnostics.Process.Start(Path.GetDirectoryName(song.Path));
        }

        #endregion

        #region Event handlers to update the UI

        /// <summary>
        /// Changes the icon of the play/pause button according to the current state.
        /// </summary>
        private void Player_StateChanged(IPlayer sender, EventArgs e)
        {
            if (sender.State == PlayerState.Playing)
                this.PlayPause.Style = this.FindResource("Pause") as Style;
            else
                this.PlayPause.Style = this.FindResource("Play") as Style;
        }

        /// <summary>
        /// Updates the TextBlocks showing information about the current song.
        /// If the window is minimized, a toaster popup is shown.
        /// </summary>
        private void Player_SongChanged(IPlayer sender, EventArgs e)
        {
            this.Dispatcher.Invoke(new Action(() =>
            {
                this.SongTitle.Text = App.This.Player.Song.Title;
                this.Artist.Text = App.This.Player.Song.Artist;
                this.Album.Text = App.This.Player.Song.Album;

                if (this.WindowState == WindowState.Minimized)
                    Toast.Instance.Show(App.This.Player.Song);
            }), new object[0]);
        }

        /// <summary>
        /// Disables/enables the previous/next buttons as necessary.
        /// Tip: For more details about the difference between SongChanged and SongIndexChanged,
        /// see IPlayer! ;-)
        /// </summary>
        private void Player_SongIndexChanged(IPlayer arg1, EventArgs arg2) {
            this.Dispatcher.Invoke(new Action(() => {
                this.UpdatePrevNext();
            }));
        }

        /// <summary>
        /// Disables/enables the previous/next buttons as necessary.
        /// </summary>
        protected void UpdatePrevNext() {
            this.Previous.IsEnabled = App.This.Player.SongIndex > 0 || App.This.Player.Repeat;
            this.Next.IsEnabled = App.This.Player.SongIndex < App.This.Player.Playlist.Count - 1 || App.This.Player.Repeat;
        }

        //Show the songs from the selected playlist
        private void Playlists_SelectionChanged(object sender, SelectionChangedEventArgs e)
        {
            if (this.PlaylistsList.SelectedItem == null)
                this.PlaylistsList.SelectedIndex = 0;
            this.LoadSongs();
        }

        #endregion

        #region ListView sorting

        //Some code from the internet for ListView sorting
        GridViewColumnHeader _lastHeaderClicked = null;
        ListSortDirection _lastDirection = ListSortDirection.Ascending;
        private void Songs_Click(object sender, RoutedEventArgs e)
        {
            GridViewColumnHeader headerClicked = e.OriginalSource as GridViewColumnHeader;
            ListSortDirection direction;

            if (headerClicked != null)
            {
                if (headerClicked.Role != GridViewColumnHeaderRole.Padding)
                {
                    if (headerClicked != _lastHeaderClicked)
                    {
                        direction = ListSortDirection.Ascending;
                    }
                    else
                    {
                        if (_lastDirection == ListSortDirection.Ascending)
                        {
                            direction = ListSortDirection.Descending;
                        }
                        else
                        {
                            direction = ListSortDirection.Ascending;
                        }
                    }

                    string header = headerClicked.Column.Header as string;
                    Sort(header, direction);

                    if (direction == ListSortDirection.Ascending)
                    {
                        headerClicked.Column.HeaderTemplate =
                          Resources["HeaderTemplateArrowUp"] as DataTemplate;
                    }
                    else
                    {
                        headerClicked.Column.HeaderTemplate =
                          Resources["HeaderTemplateArrowDown"] as DataTemplate;
                    }

                    // Remove arrow from previously sorted header
                    if (_lastHeaderClicked != null && _lastHeaderClicked != headerClicked)
                    {
                        _lastHeaderClicked.Column.HeaderTemplate = null;
                    }


                    _lastHeaderClicked = headerClicked;
                    _lastDirection = direction;
                }
            }
        }

        private void Sort(string sortBy, ListSortDirection direction)
        {
            ICollectionView dataView =
              CollectionViewSource.GetDefaultView(this.Songs.Items);

            dataView.SortDescriptions.Clear();
            SortDescription sd = new SortDescription(sortBy, direction);
            dataView.SortDescriptions.Add(sd);
            dataView.Refresh();
        }

        #endregion


    }
}
