﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Documents;
using System.Windows.Media;
using System.Windows.Shapes;
using System.Windows.Media.Animation;
using System.Windows.Input;

namespace Musiclibrary {
    /// <summary>
    /// Provides an overlay (in WPF called: Adorner) that shows a popup
    /// </summary>
    public class DialogAdorner : Adorner {
        /// <param name="adornedElement">The element that should be overlayed.</param>
        /// <param name="content">The popup content.</param>
        /// <param name="canCancel">Can the user close the popup by clicking in an empty area?</param>
        public DialogAdorner(FrameworkElement adornedElement, Popup content, bool canCancel) :
            base(adornedElement) {

            //Create control structure
            var canvas = new Canvas();
            canvas.Opacity = 0;

            var overlay = new Rectangle();
            overlay.Fill = Brushes.Black;
            overlay.Opacity = 0.5;
            overlay.HorizontalAlignment = HorizontalAlignment.Stretch;
            overlay.VerticalAlignment = VerticalAlignment.Stretch;

            var box = new Border();
            box.BorderBrush = new SolidColorBrush(Color.FromArgb(0xff, 0x6b, 0xb6, 0x22));
            box.BorderThickness = new Thickness(10);
            box.CornerRadius = new CornerRadius(10);
            box.Background = Brushes.White;
            box.Child = content;
            box.SnapsToDevicePixels = true;
            
            canvas.Children.Add(overlay);
            canvas.Children.Add(box);

            this.VisualChild = canvas;
            this.Overlay = overlay;
            this.Box = box;
            this.Content = content;

            content.Adorner = this;

            this.AddLogicalChild(this.VisualChild);
            this.AddVisualChild(this.VisualChild);

            this.UpdateDialog();

            //Reposition dialog when adorned element changes size
            adornedElement.SizeChanged += new SizeChangedEventHandler((sender, e) => 
                this.UpdateDialog());

            if (canCancel)
                overlay.MouseDown += new MouseButtonEventHandler(overlay_MouseDown);
        }

        void overlay_MouseDown(object sender, MouseButtonEventArgs e) {
            this.Hide();
        }


        public bool IsShown { get; protected set; }
        protected bool CanCancel { get; set; }
        public Popup Content { get; set; }
        protected Canvas VisualChild { get; set; }
        protected Border Box { get; set; }
        protected Rectangle Overlay { get; set; }

        /// <summary>
        /// Updates the popup's position so that it is always centered
        /// </summary>
        protected void UpdateDialog() {
            this.VisualChild.Width = ((FrameworkElement)this.AdornedElement).ActualWidth;
            this.VisualChild.Height = ((FrameworkElement)this.AdornedElement).ActualHeight;
            this.Overlay.Width = this.VisualChild.Width;
            this.Overlay.Height = this.VisualChild.Height;

            this.Box.Measure(new Size(double.PositiveInfinity, double.PositiveInfinity));
            this.Box.Arrange(new Rect());

            Canvas.SetLeft(this.Box, (this.VisualChild.Width - this.Box.ActualWidth) / 2);
            Canvas.SetTop(this.Box, (this.VisualChild.Height - this.Box.ActualHeight) / 2);
        }

        #region Show/Hide

        /// <summary>
        /// Show the popup with a fancy animation
        /// </summary>
        public void Show() {
            this.IsShown = true;

            var layer = AdornerLayer.GetAdornerLayer(this.AdornedElement);
            layer.Add(this);

            var duration = new TimeSpan(0, 0, 0, 0, 300);

            var storyboard = new Storyboard();

            var fadeIn = new DoubleAnimation();
            fadeIn.To = 1;
            fadeIn.From = 0;
            fadeIn.Duration = duration;
            storyboard.Children.Add(fadeIn);
            Storyboard.SetTarget(fadeIn, this.VisualChild);
            Storyboard.SetTargetProperty(fadeIn, new PropertyPath("(Canvas.Opacity)"));

            var slideDown = new DoubleAnimation();
            slideDown.To = Canvas.GetTop(this.Box);
            slideDown.From = -this.Box.ActualHeight;
            slideDown.Duration = duration;
            slideDown.DecelerationRatio = 1;
            storyboard.Children.Add(slideDown);
            Storyboard.SetTarget(slideDown, this.Box);
            Storyboard.SetTargetProperty(slideDown, new PropertyPath("(Canvas.Top)"));

            storyboard.Begin();
        }

        /// <summary>
        /// Hide the popup, again with a fancy effect
        /// </summary>
        public void Hide() {
            this.IsShown = false;

            var duration = new TimeSpan(0, 0, 0, 0, 300);

            var storyboard = new Storyboard();

            var fadeOut = new DoubleAnimation();
            fadeOut.To = 0;
            fadeOut.Duration = new TimeSpan(0, 0, 0, 0, 300);
            storyboard.Children.Add(fadeOut);
            Storyboard.SetTarget(fadeOut, this.VisualChild);
            Storyboard.SetTargetProperty(fadeOut, new PropertyPath("(Canvas.Opacity)"));

            storyboard.Begin();

            duration.After(new Action(() =>
                AdornerLayer.GetAdornerLayer(this.AdornedElement).Remove(this)), this.AdornedElement);
        }

        private Storyboard currentAnimation;

        /// <summary>
        /// Move to the next popup (e.g. in a wizard) with another cool effect! :D
        /// </summary>
        public void ShowNextDialog(Popup content, bool canCancel) {
            this.IsShown = true;
            var duration = new TimeSpan(0, 0, 0, 0, 300);

            if (this.currentAnimation != null) currentAnimation.Stop();

            var storyboard = new Storyboard();

            var slideOut = new DoubleAnimation();
            slideOut.To = -Canvas.GetLeft(this.Box);
            slideOut.Duration = duration;
            slideOut.AccelerationRatio = 1;
            slideOut.Duration = duration;
            storyboard.Children.Add(slideOut);
            Storyboard.SetTarget(slideOut, this.Box);
            Storyboard.SetTargetProperty(slideOut, new PropertyPath("(Canvas.Left)"));

            var slideIn = new DoubleAnimation();
            slideIn.From = this.VisualChild.Width;
            slideIn.DecelerationRatio = 1;
            slideIn.BeginTime = duration;
            slideIn.Duration = duration;
            storyboard.Children.Add(slideIn);
            Storyboard.SetTarget(slideIn, this.Box);
            Storyboard.SetTargetProperty(slideIn, new PropertyPath("(Canvas.Left)"));

            storyboard.Begin();
            this.currentAnimation = storyboard;

            duration.After(new Action(() => {
                this.Content = content;
                this.Box.Child = content;
                content.Adorner = this;
                slideIn.To = (this.Overlay.Width - this.Content.ActualWidth) / 2;

            }), this.AdornedElement);

            if (canCancel)
                this.Overlay.MouseDown += new MouseButtonEventHandler(overlay_MouseDown);
            else
                this.Overlay.MouseDown -= overlay_MouseDown;
        }

        #endregion

        #region Overrides

        protected override Size MeasureOverride(Size constraint) {
            this.VisualChild.Measure(constraint);
            return this.VisualChild.DesiredSize;
        }

        protected override Size ArrangeOverride(Size finalSize) {
            this.VisualChild.Arrange(new Rect(finalSize));
            return finalSize;
        }

        protected override System.Windows.Media.Visual GetVisualChild(int index) {
            return this.VisualChild;
        }

        protected override int VisualChildrenCount {
            get {
                return 1;
            }
        }

        #endregion
    }
}
