﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

namespace Musiclibrary {
    /// <summary>
    /// Represents a class that can play music files.
    /// Currently, there is only one implementation: WMPPlayer.
    /// </summary>
    public interface IPlayer {
        /// <summary>
        /// The playlist the player is currently going through.
        /// </summary>
        List<ISong> Playlist { get; }

        /// <summary>
        /// The index of the currently played song in the playlist.
        /// This property is needed, because the same song can be added several times to the playlist.
        /// </summary>
        int SongIndex { get; set; }

        /// <summary>
        /// Returns the currently played or paused song.
        /// </summary>
        ISong Song { get; set; }

        /// <summary>
        /// Is raised when the current song changes.
        /// </summary>
        event Action<IPlayer, EventArgs> SongChanged;

        /// <summary>
        /// Is raised if the index of the current song in the playlist is changed.
        /// This event is raised every time SongChanged is raised, when the Shuffle() function is called
        /// and when the Repeat property is changed.
        /// It is thought for enabling/disabling the previous/next song buttons.
        /// </summary>
        event Action<IPlayer, EventArgs> SongIndexChanged;

        /// <summary>
        /// The current player state: Playing, paused or finished.
        /// </summary>
        PlayerState State { get; }

        /// <summary>
        /// Is raised when the state changes.
        /// </summary>
        event Action<IPlayer, EventArgs> StateChanged;

        /// <summary>
        /// The position in the current song.
        /// </summary>
        TimeSpan Position { get; set; }

        /// <summary>
        /// The duration of the current song.
        /// </summary>
        TimeSpan Duration { get; }

        /// <summary>
        /// The current player volume, in a range from 0 to 1.
        /// </summary>
        double Volume { get; set; }

        /// <summary>
        /// Starts playing!
        /// </summary>
        void Play();

        /// <summary>
        /// Pauses the current song.
        /// </summary>
        void Pause();

        /// <summary>
        /// Jumps to the next song in the playlist. If there is no next song, the function does nothing.
        /// This behavior is changed if the Repeat property is set to true.
        /// </summary>
        void Next();

        /// <summary>
        /// Jumps to the previous song in the playlist. If there is no previous song, the function does nothing.
        /// This behavior is changed if the Repeat property is set to true.
        /// </summary>
        void Previous();

        /// <summary>
        /// Determines whether the player should repeat the current playlist all the time.
        /// </summary>
        bool Repeat { get; set; }

        /// <summary>
        /// Randomly changes the order of the items in the current playlist.
        /// </summary>
        void Shuffle();
    }

    public enum PlayerState {
        Playing,
        Paused,
        Finished
    }
}
