﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Net;
using System.Net.Sockets;
using System.Threading;
using System.IO;

namespace Musiclibrary {
    /// <summary>
    /// Represents a connection from a music library server to a client.
    /// Provides functions to process the client's requests.
    /// </summary>
    public class ClientConnection {
        public ClientConnection(Socket socket) {
            this.Socket = socket;

            this.Thread = new Thread(new ThreadStart(this.ReplyToCommands));
            this.Thread.Start();

            this.IsConnected = true;
        }

        /// <summary>
        /// Sends a disconnect command to the client and closes the connection.
        /// The object cannot be reused after calling this function.
        /// </summary>
        public void Disconnect() {
            this.Thread.Abort();
            this.SendCommand("Disconnect!");
            this.Socket.Close();
            this.IsConnected = false;
        }

        public bool IsConnected { get; protected set; }

        protected Thread Thread { get; set; }
        protected Socket Socket { get; set; }

        protected void ReplyToCommands() {
            while (true) {
                try {
                    var command = this.ReceiveCommand();

                    if (command.StartsWith("Give me all songs!"))
                        this.SendSongList(App.This.Library.Songs);
                    else if (command.StartsWith("The playlists, please!"))
                        this.SendPlaylists(App.This.Library.Playlists);
                    else if (command.StartsWith("I'd like to have the songs from playlist #")) {
                        int playlistId;
                        if (int.TryParse(command.Substring(42), out playlistId)) { //42!!!
                            var result = from i in App.This.Library.Playlists where i.Id == playlistId select i;

                            if (result.Count() > 0) {
                                var playlist = result.First();

                                this.SendSongList(playlist.Songs.Cast<Song>());
                            }
                        }
                    } else if (command.StartsWith("Send me song #")) {
                        int songId;
                        if (int.TryParse(command.Substring(14), out songId)) {
                            var result = from i in App.This.Library.Songs where i.Id == songId select i;

                            if (result.Count() > 0) {
                                var song = result.First();

                                this.SendSongData(song);
                            }
                        }
                    } else if (command == "Disconnect!") {
                        this.IsConnected = false;
                        this.Socket.Close();
                        return;
                    }
                } 
                catch (ThreadAbortException) {
                    return; //Get out of function when thread should be cancelled
                } 
                catch { 
                    //Ignore eventual socket errors
                }
            }
        }

        /// <summary>
        /// Sends a command to the client.
        /// </summary>
        protected void SendCommand(string command) {
            //Command delimiter is necessary because a stream connection is used
            var data = UTF8Encoding.UTF8.GetBytes(command.Escape('\0') + "\0");

            for (var i = 0; i < data.Length; i += 1024) {
                int len;
                if (data.Length - i < 1024)
                    len = data.Length - i;
                else
                    len = 1024;

                this.Socket.Send(data, i, len, SocketFlags.None);
            }
        }

        private Queue<string> CommandQueue = new Queue<string>();
        private string CurrentCommand = "";

        /// <summary>
        /// Returns the oldest command received that has not been processed yet.
        /// </summary>
        protected string ReceiveCommand() {
            while (this.CommandQueue.Count == 0) {
                //Receive data
                var data = new byte[1024];
                var len = this.Socket.Receive(data);
                var message = UTF8Encoding.UTF8.GetString(data, 0, len);

                //Split message into commands
                bool complete;
                var commands = (this.CurrentCommand + message).SplitEscaped('\0', out complete);

                //If last command is not finished, store it in the CurrentCommand variable
                if (!complete) {
                    this.CurrentCommand += commands[commands.Count - 1];
                    commands.RemoveAt(commands.Count - 1);
                } else
                    this.CurrentCommand = "";

                //Put commands in the queue
                foreach (var command in commands)
                    this.CommandQueue.Enqueue(command);
            }

            return this.CommandQueue.Dequeue();
        }

        /// <summary>
        /// Sends the metadata from one song to the client.
        /// </summary>
        protected void SendSongMetadata(Song song) {
            //Format: #[Id];[Title];[Arist];[Album];[Genre];[Year];
            var message = "#" + song.Id.ToString() + ";" + 
                song.Title.Escape(';') + ";" +
                song.Artist.Escape(';') + ";" +
                song.Album.Escape(';') + ";" +
                song.Genre.Escape(';') + ";";

            if (song.Year.HasValue)
                message += song.Year.Value.ToString() + ";";
            else
                message += "null;";

            this.SendCommand(message);
        }

        /// <summary>
        /// Sends the metadata for multiple songs to the client.
        /// </summary
        protected void SendSongList(IEnumerable<Song> songs) {
            foreach (var song in songs)
                this.SendSongMetadata(song);

            this.SendCommand("Okay, that's it!");
        }

        /// <summary>
        /// Sends a list of playlists to the client.
        /// </summary>
        protected void SendPlaylists(IEnumerable<Playlist> playlists) {
            //Format: [Id] [Name]
            foreach (var playlist in playlists)
                this.SendCommand(playlist.Id.ToString() + " " + playlist.Name);

            this.SendCommand("Okay, that's it!");
        }

        /// <summary>
        /// Sends a song's data using a second connection.
        /// </summary>
        protected void SendSongData(Song song) {
            //Connect on port 16791
            var dataServer = new Socket(AddressFamily.InterNetwork, SocketType.Stream, ProtocolType.Tcp);
            dataServer.Bind(new IPEndPoint(IPAddress.Any, 16791));
            dataServer.Listen(1);
            var dataSocket = dataServer.Accept();

            //Send data
            var stream = new FileInfo(song.Path).OpenRead();
            while (stream.Position < stream.Length) {
                var data = new byte[1024];
                var len = stream.Read(data, 0, 1024);
                dataSocket.Send(data, len, SocketFlags.None);
            }

            //Disconnect
            dataSocket.Close();
            dataServer.Close();
        }

        
    }
}
