﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Runtime.Remoting;
using System.Runtime.Remoting.Channels;
using System.Runtime.Remoting.Channels.Tcp;
using System.Reflection;
using System.Net;
using System.Net.Sockets;
using System.Windows;
using System.Threading;
using System.IO;

//Sharing ports:
//16789: Broadcast/discovery
//16790: Main connection
//16791: Data transfer

namespace Musiclibrary {
    /// <summary>
    /// Represents a remote library.
    /// </summary>
    public class RemoteLibrary : ILibrary {
        #region Library discovery

        /// <summary>
        /// Looks for other libraries.
        /// </summary>
        public static List<RemoteLibraryInfo> Discover() {
            var libraries = new List<RemoteLibraryInfo>();

            try {
                //Send broadcast
                var localEP = SendBroadcast("Hello! Any music library there?");

                //Create socket to receive replies
                var socket = new Socket(AddressFamily.InterNetwork, SocketType.Dgram, ProtocolType.Udp);
                socket.Bind(localEP);

                //Set receive timeout to make this method not endless-blocking
                socket.ReceiveTimeout = 1000;

                //Receive as many replies as possible
                while (true) {
                    //Receive reply
                    byte[] buffer = new byte[1024];
                    EndPoint remoteEP = new IPEndPoint(IPAddress.Loopback, 11111);
                    socket.ReceiveFrom(buffer, ref remoteEP);
                    var message = UTF8Encoding.UTF8.GetString(buffer);

                    //Is reply from other music library ?
                    if (message.StartsWith("Yoo! I am ")) {
                        //Add library to list
                        var name = message.Substring(10).Trim((char)0);
                        remoteEP = new IPEndPoint((remoteEP as IPEndPoint).Address, 16790);

                        //Don't connect to own library
                        if ((remoteEP as IPEndPoint).Address.ToString() != 
                                (localEP as IPEndPoint).Address.ToString())
                            libraries.Add(new RemoteLibraryInfo() {
                                Name = name,
                                EndPoint = remoteEP
                            });
                    }
                }
            } catch { }

            return libraries;
        }

        /// <summary>
        /// Sends a UDP broadcast on port 16789
        /// </summary>
        /// <returns>The local end point.</returns>
        private static EndPoint SendBroadcast(string message) {
            byte[] buffer = new byte[1024];
            var socket = new Socket(AddressFamily.InterNetwork, SocketType.Dgram, ProtocolType.Udp);
            socket.SetSocketOption(SocketOptionLevel.Socket, SocketOptionName.Broadcast, 1);

            socket.Connect(new IPEndPoint(IPAddress.Broadcast, 16789));
            socket.Send(System.Text.UTF8Encoding.UTF8.GetBytes(message));

            var ep = socket.LocalEndPoint;

            socket.Close();

            return ep;
        }

        #endregion

        protected Socket Socket;
        protected Thread ReceiveThread;

        /// <summary>
        /// Connects to a remote library at the specified EndPoint.
        /// </summary>
        public RemoteLibrary(EndPoint remoteEndPoint) {
            this.Socket = new Socket(AddressFamily.InterNetwork, SocketType.Stream, ProtocolType.Tcp);
            this.Socket.Connect(remoteEndPoint);
            this.IsConnected = true;

            this.ReceiveThread = new Thread(new ThreadStart(ReceiveCommands));
            this.ReceiveThread.Start();
        }

        public bool IsConnected { get; protected set; }

        public event EventHandler Disconnected;

        /// <summary>
        /// Gets all songs from the playlist with the specified id.
        /// </summary>
        /// <param name="playlistId"></param>
        /// <returns></returns>
        public IEnumerable<ISong> GetPlaylistSongs(int playlistId) {
            this.SendCommand("I'd like to have the songs from playlist #" + playlistId.ToString());
            return this.ReceiveSongList().Cast<ISong>();
        }

        /// <summary>
        /// Loads a song from the remote library and returns the path to a temp file containing the song.
        /// </summary>
        /// <returns></returns>
        public string LoadSong(int songId) {
            this.SendCommand("Send me song #" + songId.ToString());

            var tempFile = Extensions.GetTemporaryFile("mp3");
            var stream = new FileStream(tempFile, FileMode.Create);
            this.ReceiveSongData(stream);
            stream.Close();

            return tempFile;
        }

        /// <summary>
        /// Sends a disconnect command to the server and closes the connection.
        /// </summary>
        public void Disconnect() {
            this.SendCommand("Disconnect!");
            this.Socket.Close();

            if (this.Disconnected != null)
                this.Disconnected(this, EventArgs.Empty);
        }

        private Queue<string> CommandQueue = new Queue<string>();
        private string CurrentCommand = "";

        /// <summary>
        /// Receives all commands coming from the server.
        /// </summary>
        protected void ReceiveCommands() {
            while (true) {
                try {
                    //Receive data
                    var data = new byte[1024];
                    var len = this.Socket.Receive(data);
                    var message = UTF8Encoding.UTF8.GetString(data, 0, len);

                    //Split message into commands
                    bool complete;
                    var commands = (this.CurrentCommand + message).SplitEscaped('\0', out complete);

                    //If last command is not finished, store it in the CurrentCommand variable
                    if (!complete) {
                        this.CurrentCommand += commands[commands.Count - 1];
                        commands.RemoveAt(commands.Count - 1);
                    } else
                        this.CurrentCommand = "";

                    //Put commands in the queue
                    foreach (var command in commands)
                        this.CommandQueue.Enqueue(command);

                    //Process disconnect command, the only true command the client receives.
                    if (commands.Contains("Disconnect!")) {
                        this.IsConnected = false;
                        App.This.RemoteLibraries.Remove(this.Socket.RemoteEndPoint);
                        if (this.Disconnected != null)
                            this.Disconnected(this, EventArgs.Empty);
                        return;
                    }
                } catch { }
            }
        }

        /// <summary>
        /// Returns the oldest command received. If there is no command in the queue, wait until there is one.
        /// </summary>
        /// <returns></returns>
        protected string ReceiveCommand() {
            for (var i = 0; i < 100 && this.CommandQueue.Count == 0; ++i)
                Thread.Sleep(20);

            if (this.CommandQueue.Count == 0)
                throw new Exception("No reply from the server");

            return this.CommandQueue.Dequeue();
        }

        /// <summary>
        /// Sends a command to the server.
        /// </summary>
        protected void SendCommand(string command) {
            //Command delimiter is necessary because a stream connection is used
            var data = UTF8Encoding.UTF8.GetBytes(command.Escape('\0') + "\0");

            for (var i = 0; i < data.Length; i += 1024) {
                int len;
                if (data.Length - i < 1024)
                    len = data.Length - i;
                else
                    len = 1024;

                this.Socket.Send(data, i, len, SocketFlags.None);
            }
        }

        /// <summary>
        /// Parses a song's metadata from a message from the server and returns a RemoteSong.
        /// </summary>
        /// <returns></returns>
        protected RemoteSong ParseSong(string message) {
            var data = message.Substring(1).SplitEscaped(';', out (new bool[1])[0]);
            var song = new RemoteSong(this);

            song.Id = int.Parse(data[0]);
            song.Title = data[1];
            song.Artist = data[2];
            song.Album = data[3];
            song.Genre = data[4];

            int year;
            if (int.TryParse(data[5], out year))
                song.Year = year;

            return song;
        }

        /// <summary>
        /// Receives a list of song metadata.
        /// </summary>
        /// <returns></returns>
        protected List<RemoteSong> ReceiveSongList() {
            var songs = new List<RemoteSong>();

            for (string message = this.ReceiveCommand(); message != "Okay, that's it!"; message = this.ReceiveCommand())
                songs.Add(this.ParseSong(message));

            return songs;
        }


        /// <summary>
        /// Receives a playlist-list.
        /// </summary>
        /// <returns></returns>
        protected List<RemotePlaylist> ReceivePlaylists() {
            var playlists = new List<RemotePlaylist>();

            for (string message = this.ReceiveCommand(); message != "Okay, that's it!"; message = this.ReceiveCommand()) {
                var playlist = new RemotePlaylist(this);

                var splitPos = message.IndexOf(' ');
                playlist.Id = int.Parse(message.Substring(0, splitPos));
                playlist.Name = message.Substring(splitPos + 1);

                playlists.Add(playlist);
            }

            return playlists;
        }

        /// <summary>
        /// Receives a song's data using a second connection.
        /// </summary>
        /// <param name="target">The stream to which the data should be written.</param>
        protected void ReceiveSongData(Stream target) {
            var dataSocket = new Socket(AddressFamily.InterNetwork, SocketType.Stream, ProtocolType.Tcp);
            dataSocket.Connect(new IPEndPoint((this.Socket.RemoteEndPoint as IPEndPoint).Address, 16791));
            
            while (true) {
                byte[] data = new byte[1024];
                int len;

                try {
                    len = dataSocket.Receive(data);
                } catch (ObjectDisposedException) {
                    dataSocket.Close();
                    return;
                }

                if (!dataSocket.Connected || len == 0) {
                    dataSocket.Close();
                    return;
                }
                else
                    target.Write(data, 0, len);
            }
        }

        #region ILibrary Member

        public string Name {
            get;
            set;
        }

        public IEnumerable<ISong> Songs {
            get {
                this.SendCommand("Give me all songs!");
                return this.ReceiveSongList().Cast<ISong>();
            }
        }

        public IEnumerable<IPlaylist> Playlists {
            get {
                this.SendCommand("The playlists, please!");
                return this.ReceivePlaylists().Cast<IPlaylist>();
            }
        }

        #endregion
    }
}