﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Windows.Threading;
using System.Timers;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Media;
using System.IO;

namespace Musiclibrary {
    /// <summary>
    /// This class contains several extension methods and other helper functions that 
    /// have absolutely nothing in common that would justify that they are all grouped in this class.
    /// </summary>
    public static class Extensions {

        /// <summary>
        /// Does something after a given time span.
        /// </summary>
        /// <param name="invokeTarget">A target to invoke the action.</param>
        public static void After(this TimeSpan time, Action action, DispatcherObject invokeTarget) {
            var timer = new Timer(time.TotalMilliseconds);
            timer.Elapsed += new ElapsedEventHandler((sender, e) => {
                timer.Stop();
                invokeTarget.Dispatcher.Invoke(action, new object[0]);
            });
            timer.Start();
        }

        /// <summary>
        /// Does something after a given time span. 
        /// </summary>
        public static void After(this TimeSpan time, Action action) {
            var timer = new Timer(time.TotalMilliseconds);
            timer.Elapsed += new ElapsedEventHandler((sender, e) => {
                timer.Stop();
                action();
            });
            timer.Start();
        }

        /// <summary>
        /// Imagine, you're in an event handler raised by an ItemsControl or one of its visual children.
        /// Now you want to find out which item has been clicked? Then this function is right for you!
        /// </summary>
        /// <typeparam name="TItem">The type of the item.</typeparam>
        /// <param name="container">The ItemsControl.</param>
        /// <param name="sender">The event handler's sender parameter.</param>
        /// <returns></returns>
        public static TItem GetItemFromEventSender<TItem>(this ItemsControl container, object sender) where TItem : class {
            UIElement element = sender as UIElement;

            object data = DependencyProperty.UnsetValue;
            while (data == DependencyProperty.UnsetValue) {
                data = container.ItemContainerGenerator.ItemFromContainer(element);

                if (data == DependencyProperty.UnsetValue)
                    element = VisualTreeHelper.GetParent(element) as UIElement;

                if (element == container)
                    return null;
            }

            if (data != DependencyProperty.UnsetValue && data is TItem)
                return data as TItem;
            else
                return null;
        }

        /// <summary>
        /// Formats a time span like that: MM:SS
        /// </summary>
        /// <returns></returns>
        public static string ToShortString(this TimeSpan time) {
            return time.Minutes.ToString("00") + ":" + time.Seconds.ToString("00");
        }

        /// <summary>
        /// Determines whether a list is empty.
        /// </summary>
        public static bool IsEmpty<T>(this IEnumerable<T> list)
        {
            foreach (var item in list)
                return false;

            return true;
        }

        /// <summary>
        /// Filters a list of songs by a specified search string.
        /// </summary>
        public static IEnumerable<ISong> Filter(this IEnumerable<ISong> list, string search)
        {
            search = search.ToLower().Trim();

            return from i in list
                   where
                    i.Title.ToLower().Contains(search) ||
                    i.Artist.ToLower().Contains(search) ||
                    i.Album.ToLower().Contains(search) ||
                    i.Genre.ToLower().Contains(search) ||
                    i.Year.ToString().Contains(search)
                   select i;
        }

        /// <summary>
        /// Removes all invalid chars from a future file name.
        /// </summary>
        /// <param name="path"></param>
        /// <returns></returns>
        public static string RemoveInvalidPathChars(this string path)
        {
            var invalidChars = Path.GetInvalidFileNameChars().Union(Path.GetInvalidPathChars());

            foreach (var chr in invalidChars)
                path = path.Replace(chr.ToString(), "");

            return path;
        }

        /// <summary>
        /// Compares two files regarding every byte, if necessary.
        /// </summary>
        /// <returns></returns>
        public static bool Compare(this FileInfo file1, FileInfo file2)
        {
            byte[] buffer1 = new byte[2048];
            byte[] buffer2 = new byte[2048];


            // Determine if the same file was referenced two times.
            if (file1 == file2) return true;

            // Check the file sizes. If they are not the same, the files 
            // are not the same.
            if (file1.Length != file2.Length) return false;

            // Open the two files.
            var fs1 = file1.OpenRead();
            var fs2 = file2.OpenRead();

            // Read and compare a byte from each file until either a
            // non-matching set of bytes is found or until the end of
            // file1 is reached.
            while (fs1.Position < fs1.Length)
            {
                // Read 2048 bytes from each file.
                var count = fs1.Read(buffer1, 0, 2048);
                fs2.Read(buffer2, 0, 2048);

                for (var i = 0; i < count; ++i)
                    if (buffer1[i] != buffer2[i])
                    {
                        fs1.Close();
                        fs2.Close();
                        return false;
                    }
            }

            fs1.Close();
            fs2.Close();

            return true;
        }

        /// <summary>
        /// Creates a temporary file with the specified extension.
        /// </summary>
        /// <returns>The path to the file.</returns>
        public static string GetTemporaryFile(string extension) {
            string response = string.Empty;

            if (!extension.StartsWith("."))
                extension = "." + extension;

            response = Path.GetTempPath() + Guid.NewGuid().ToString() + extension;

            return response;
        }

        /// <summary>
        /// Escapes all specified chars with a backslash.
        /// </summary>
        public static string Escape(this string text, params char[] charsToEscape) {
            text = text.Replace("\\", "\\\\");
            foreach (var chr in charsToEscape)
                text = text.Replace(chr.ToString(), "\\" + chr.ToString());
            return text;
        }

        /// <summary>
        /// Splits a string seperated by a specified char ignoring escaped occurences 
        /// of this char.
        /// </summary>
        /// <param name="text"></param>
        /// <param name="splitChar"></param>
        /// <param name="complete">Determines whether a splitChar occurs at the end of the text.</param>
        /// <returns></returns>
        public static List<string> SplitEscaped(this string text, char splitChar, out bool complete) {
            var parts = new List<string>();

            var part = "";
            bool escaped = false;
            for (var i = 0; i < text.Length; ++i) {
                if (!escaped) {
                    if (text[i] == splitChar) {
                        parts.Add(part);
                        part = "";
                    } else if (text[i] == '\\')
                        escaped = true;
                    else
                        part += text[i];
                } else {
                    part += text[i];
                    escaped = false;
                }
            }

            if (text[text.Length - 1] != splitChar) {
                parts.Add(part);
                complete = false;
            } else
                complete = true;

            return parts;
        }

        /// <summary>
        /// Shows a message box with an error icon.
        /// </summary>
        public static void ShowErrorBox(string message) {
            MessageBox.Show(message, "Musiclibrary", MessageBoxButton.OK, MessageBoxImage.Error);
        }
    }
}
