﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.IO;
using System.Windows.Threading;
namespace Musiclibrary
{
    /// <summary>
    /// Imports the files and/or directories specified in the FilesAndDirs property.
    /// </summary>
    public class Importer : Worker
    {
        /// <summary>
        /// 
        /// </summary>
        /// <param name="invokeTarget">A DispatcherObject which is used for invoking when any events are raised.</param>
        public Importer(DispatcherObject invokeTarget)
            : base(invokeTarget)
        {
            this.FilesAndDirs = new List<FileSystemInfo>();
        }

        public List<FileSystemInfo> FilesAndDirs { get; private set; }
        protected List<FileInfo> Files { get; set; }

        protected override object DoWork()
        {
            //Collect all mp3 files from the specified directories
            this.CurrentState = "Collecting files...";
            this.Progress = null;
            this.Collect();

            if (this.CancelPending) return null;

            var alreadExists = new List<FileInfo>();
            var failed = new Dictionary<FileInfo, Exception>();

            var i = 1;
            foreach (var file in this.Files)
            {
                this.CurrentState = "Importing \"" + file.Name + "\"...";

                try
                {
                    //Load the ID3 data. This is necessary already here, because the next step is to validate
                    //whether the song is already in the library. If the library is organized, this cannot be
                    //determined by the path of the file, but only by the reorganized path which is computed
                    //from the metadata
                    var info = MetadataBase.LoadMetadata(file.FullName);
                    var song = info.GetSong();

                    //As I said, check if the song is already in the library
                    if (this.Validate(file, song))
                       this.Import(file, song);
                    else
                       alreadExists.Add(file);
                }
                catch (Exception ex)
                {
                    failed.Add(file, ex);
                }
               
                this.Progress = ((double)i++ / (double)this.Files.Count);

                if (this.CancelPending) break;
            }

            try {
                App.This.Library.SubmitChanges();
            } catch (Exception ex) {
                failed.Clear();
                failed.Add(new FileInfo("General error"), ex);
            }

            if (!this.CancelPending)
                return new ImportResult(alreadExists, failed);
            else
                return null;
        }

        /// <summary>
        /// Takes the FilesAndDirs property and looks for all mp3 files in the directories.
        /// Then the resulting files are saved in the Files property.
        /// </summary>
        protected void Collect()
        {
            this.Files = new List<FileInfo>();

            foreach (var fsInfo in this.FilesAndDirs)
            {
                if (fsInfo is FileInfo)
                    this.Files.Add(fsInfo as FileInfo);
                else if (fsInfo is DirectoryInfo)
                    this.Collect(fsInfo as DirectoryInfo);

                if (this.CancelPending) return;
            }
        }

        /// <summary>
        /// Recursive function to collect mp3 files from a directory.
        /// </summary>
        protected void Collect(DirectoryInfo dir)
        {
            if (this.CancelPending) return;

            foreach (var file in dir.GetFiles("*.mp3"))
                this.Files.Add(file);

            foreach (var subDir in dir.GetDirectories())
                this.Collect(subDir);
        }

        /// <summary>
        /// Checks whether a file is already in the database.
        /// </summary>
        /// <returns></returns>
        protected bool Validate(FileInfo file, Song song)
        {
            string path;

            //Determine path of the file. If the library is organized, find out 
            //the file's target path
            if (App.This.Library.Organize)
                path = song.GetPath(App.This.Library.Location, App.This.Library.Schema, false);
            else
                path = file.FullName;

            //Look for path in database
            var pathExists = (from i in App.This.Library.Songs 
                              where i.Path == path
                              select true).Count() > 0;

            //If library is organized, validate if the files really are the same.
            if (App.This.Library.Organize && pathExists)
                return !file.Compare(new FileInfo(path));
            else
                return !pathExists;
        }

        /// <summary>
        /// Imports a file into the database, and, if the user has chosen this option,
        /// automatically moves the file to the proper place.
        /// </summary>
        protected void Import(FileInfo file, Song song)
        {
            if (App.This.Library.Organize) {
                var path = song.GetPath(App.This.Library.Location, App.This.Library.Schema, true);
                song.Path = path;

                if (path != file.FullName) {
                    var pathDir = Path.GetDirectoryName(path);
                    if (!Directory.Exists(pathDir))
                        Directory.CreateDirectory(pathDir);
                    File.Copy(file.FullName, path);
                }
            }

            App.This.Library.Songs.InsertOnSubmit(song);
        }
    }

    public class ImportResult {
        public ImportResult(List<FileInfo> alreadyExists, Dictionary<FileInfo, Exception> failed)
        {
            this.AlreadyExists = alreadyExists;
            this.Failed = failed;
        }

        public List<FileInfo> AlreadyExists { get; private set; }
        public Dictionary<FileInfo, Exception> Failed { get; private set; }
    }
}
