﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading;
using System.Windows.Threading;

namespace Musiclibrary
{
    /// <summary>
    /// Provides functionality to report progress and the current status from a job
    /// </summary>
    public abstract class Worker
    {
        /// <param name="invokeTarget">The target to invoke when any events are raised.</param>
        public Worker(DispatcherObject invokeTarget)
        {
            this.Dispatcher = invokeTarget.Dispatcher;
        }

        private Dispatcher Dispatcher { get; set; }
        private Thread Thread { get; set; }
        protected bool CancelPending { get; private set; }

        /// <summary>
        /// The current progress in a range from 0 to 1. If null, the IsIndeterminate property on the ProgressBar should be set
        /// </summary>
        private double? _progress;
        public double? Progress {
            get {return this._progress;}
            set {
                this._progress = value;
                this.OnProgressChanged();
            }
        }

        /// <summary>
        /// The current state of the job.
        /// </summary>
        private string _currentState;
        public string CurrentState {
            get {return this._currentState;}
            set {
                this._currentState = value;
                this.OnCurrentStateChanged();
            }
        }

        protected abstract object DoWork();

        /// <summary>
        /// Begin to do the job in a background thread.
        /// </summary>
        public void Begin()
        {
            this.Thread = new Thread(() =>
            {
                var result = this.DoWork();

                if (this.CancelPending)
                    this.OnCancelled();
                else
                    this.OnDone(result);
            });
            this.Thread.Start();
        }

        /// <summary>
        /// Cancel the job as soon as possible.
        /// </summary>
        public void Cancel()
        {
            this.CancelPending = true;
        }

        protected void OnProgressChanged()
        {
            if (this.ProgressChanged != null)
                this.Dispatcher.Invoke(new Action(() => 
                    this.ProgressChanged(this, EventArgs.Empty)), new object[0]);
        }

        protected void OnCurrentStateChanged()
        {
            if (this.CurrentStateChanged != null)
                this.Dispatcher.Invoke(new Action(() => 
                    this.CurrentStateChanged(this, EventArgs.Empty)), new object[0]);
        }

        protected void OnDone(object result)
        {
            if (this.Done != null)
                this.Dispatcher.Invoke(new Action(() => 
                    this.Done(this, new WorkerCompletedEventArgs(result))), new object[0]);
        }

        protected void OnCancelled()
        {
            if (this.Cancelled != null)
                this.Dispatcher.Invoke(new Action(() =>
                    this.Cancelled(this, EventArgs.Empty)), new object[0]);
        }

        public event EventHandler<EventArgs> ProgressChanged;
        public event EventHandler<EventArgs> CurrentStateChanged;
        public event EventHandler<WorkerCompletedEventArgs> Done;
        public event EventHandler<EventArgs> Cancelled;
    }

    public class WorkerCompletedEventArgs : EventArgs {
        public WorkerCompletedEventArgs(object result) {
            this.Result = result;
        }

        public object Result {get; private set; }
    }
}
